from sqlalchemy.orm import Session
from tasteful.repositories.base_repository import BaseRepository


class SqlAlchemyRepository(BaseRepository):
    """Declare an SqlAlchemyRepository."""

    def __init__(self, session: Session) -> None:
        self.session = session



from typing import List, Type

from dependency_injector import providers
from fastapi import FastAPI, Security

from tasteful.base_flavor import BaseFlavor
from tasteful.containers.tasteful_container import TastefulContainer


class TastefulApp:
    def __init__(
        self,
        title: str,
        version: str,
        flavors: List[Type[BaseFlavor]],
        authentication_backends: list[Security],  # type: ignore
    ):
        self.app = FastAPI(
            title=title, version=version, dependencies=authentication_backends
        )
        self.container = TastefulContainer()
        self.app.container = self.container  # type: ignore
        self.flavors = flavors
        self.register_flavors()

    def register_flavors(self) -> None:
        """Register all flavors with the app."""
        for flavor_class in self.flavors:
            self.container.flavors.override(
                providers.Singleton(
                    flavor_class,
                    name=flavor_class.name,
                    prefix=flavor_class.prefix,
                    tags=flavor_class.tags,
                    services=flavor_class.services,
                    repository=flavor_class.repository,
                )
            )
            injected_flavor = self.container.flavors()
            self.app.include_router(injected_flavor.router)



"""Decorators for authentication and authorization in Tasteful applications."""

from functools import wraps
from typing import Callable


def public(func: Callable) -> Callable:
    """Mark a route as public, bypassing authentication requirements."""
    setattr(func, "is_public_route", True)

    @wraps(func)
    async def wrapper(*args, **kwargs):
        return await func(*args, **kwargs)

    return wrapper



import re
import inspect

from typing import Any, Type, Union

from dependency_injector import providers
from fastapi import APIRouter

from tasteful.containers.base_flavor_container import BaseFlavorContainer

from ..repositories.base_repository import BaseRepository
from .base_service import BaseService


class BaseFlavor:
    """Base class for creating API flavors that handle routing and endpoint registration.

    A flavor is a collection of related endpoints grouped under a common prefix and tags.
    """

    name: str
    prefix: str
    tags: list[str]
    services: dict[str, Type]
    repository: Union[Type[BaseRepository], None]

    def __init__(
        self,
        name: str = "",
        prefix: str = "",
        tags: list[str] = [],
        services: list[Type[BaseService]] = [],
        repository: Union[Type[BaseRepository], None] = None,
    ):
        """Initialize a new flavor instance."""
        self.name = name
        self.prefix = prefix
        self.tags = tags
        self.services = services
        self.repository = repository
        self.router = APIRouter(tags=self.tags, prefix=self.prefix)
        self.container = BaseFlavorContainer()

        self._register_services()
        self._router_from_controller()

    def _convert_class_name_to_snake_case(self, class_name: str) -> str:
        s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", class_name)
        return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()

    def _register_services(self) -> None:
        """Register a service with the flavor's container."""
        for service_class in self.services:
            self.container.services_factory.override(providers.Factory(service_class))
            service_attr_name = self._convert_class_name_to_snake_case(
                service_class.__name__
            )
            setattr(self, service_attr_name, service_class())

    def _router_from_controller(self, **defaults_route_args: Any) -> None:
        """Build a router from a controller instance annotated endpoints).

        Args:
        ----
            defaults_route_args: Default arguments to pass to all routes

        Raises:
        ------
            ValueError: If no endpoints are found in the controller

        """
        # Find all methods that have endpoint definitions attached
        members = inspect.getmembers(
            self, lambda x: hasattr(x, "__endpoint_definitions__")
        )

        # Register each endpoint with the router
        for _, endpoint in members:
            for endpoint_definition in getattr(endpoint, "__endpoint_definitions__"):
                kwargs = {**defaults_route_args, **endpoint_definition.kwargs}

                self.router.add_api_route(
                    endpoint_definition.path,
                    endpoint,
                    methods=[endpoint_definition.method],
                    **kwargs,
                )



from dependency_injector import containers, providers


# https://python-dependency-injector.ets-labs.org/providers/inject_self.html
class ServiceDispatcher:
    """Service dispatcher for base flavor container."""

    def __init__(self, container: containers.Container):
        self.container = container

    def get_services(self):
        """Get all factory services from the container."""
        for provider in self.container.traverse(types=[providers.Factory]):
            yield provider()


class RepositoryDispatcher:
    def __init__(self, container: containers.Container) -> None:
        self.container = container

    def get_repository(self):
        """Get the repository."""
        for provider in self.container.traverse(types=[providers.Factory]):
            yield provider()


class BaseFlavorContainer(containers.DeclarativeContainer):
    """Container for base flavor dependencies."""

    __self__ = providers.Self()

    services_factory = providers.Factory()
    repository_factory = providers.Factory()

    services = providers.Singleton(ServiceDispatcher, __self__)
    repository = providers.Singleton(RepositoryDispatcher, __self__)



from dependency_injector import containers, providers
from tasteful.base_flavor import BaseFlavor


class TastefulContainer(containers.DeclarativeContainer):
    """Declares Tasteful Container."""

    # Flavors related
    flavors = providers.Singleton(BaseFlavor)
